# -*- coding: utf-8 -*-
#
# FinLogic - Brazilian Company Financial Data
# https://github.com/crdcj/FinLogic
#
# Copyright 2022 Carlos Carvalho
#
from . import config  # noqa
from .database import update_database, search_company, database_info
from .company import Company


__version__ = "0.3.10"
__author__ = "Carlos Carvalho"

__all__ = [
    "Company",
    "update_database",
    "search_company",
    "database_info",
]



"""This module contains the Company class, which provides methods for obtaining
financial reports and calculating financial indicators for a company. Users can
set the accounting method, unit, tax rate, and language for the company object.

Classes:
    Company: Represents a company with its financial information and allows
    users to generate financial reports and indicators.

Abreviations used in code:
    dfi = input dataframe
    dfo = output dataframe

RuntimeWarning:
    Pandas query + numexpr module -> Engine has switched to 'python' because
    numexpr does not support extension array dtypes (category not supported
    yet). Please set your engine to python manually. That means that the query
    method has to use engine='python' to work with category dtype. N.B. Only
    FinLogic Dataframe uses category dtype for efficiency.

"""
from typing import Literal
import numpy as np
import pandas as pd
from . import config as cf


class Company:
    """A class to represent a company financial data.

    This class provides methods to create financial reports and to calculate
    financial indicators based on a company's accounting data. The class also
    has an AI generated dictionary to translate from Portuguese to English.

    Methods:
        report: Creates a financial report for the company.
        custom_report: Creates a custom financial report for the company.
        indicators: Calculates the financial indicators of the company.

    Raises:
        ValueError: If the input arguments are invalid.
    """

    def __init__(
        self,
        identifier: int | str,
        acc_method: Literal["consolidated", "separate"] = "consolidated",
        acc_unit: int | float | str = 1,
        tax_rate: float = 0.34,
        language: str = "english",
    ):
        """Initializes a new instance of the Company class."""
        self._initialized = False
        self.identifier = identifier
        self.acc_method = acc_method
        self.acc_unit = acc_unit
        self.tax_rate = tax_rate
        self.language = language
        self._initialized = True
        # Only set company dataframe after identifier, acc_method and acc_unit are set
        self._set_co_df()

    @property
    def identifier(self) -> int | str:
        """Set a unique identifier to select the company in FinLogic Database.

        This method sets the company's CVM and fiscal ID based on a given
        identifier. The identifier can be either the CVM ID or the Fiscal ID
        (CNPJ). If the identifier is not found in the database, a KeyError is
        raised.

        Args:
            identifier: A unique identifier to select a company in FinLogic
                Database. Both CVM ID or Fiscal ID can be used. CVM ID
                (regulator number) must be an integer. Fiscal ID must be a
                string in 'XX.XXX.XXX/XXXX-XX' format.

        Returns:
            None

        Raises:
            KeyError: If the given identifier isn't found in Finlogic Database.
        """
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: int | str):
        # Create custom data frame for ID selection
        df = pd.read_pickle(cf.FINLOGIC_DF_PATH)[
            ["co_id", "co_fiscal_id"]
        ].drop_duplicates(ignore_index=True)
        if identifier in df["co_id"].to_list():
            self._co_id = identifier
            mask = df["co_id"] == identifier
            self._co_fiscal_id = df.loc[mask, "co_fiscal_id"].item()
        elif identifier in df["co_fiscal_id"].to_list():
            self._co_fiscal_id = identifier
            mask = df["co_fiscal_id"] == identifier
            self._co_id = df.loc[mask, "co_id"].item()
        else:
            raise KeyError("Company 'identifier' not found in FinLogic Database")
        self._identifier = identifier
        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_co_df()

    @property
    def acc_method(self) -> Literal["consolidated", "separate"]:
        """Gets or sets the accounting method for registering investments in
        subsidiaries.

        The "acc_method" must be "consolidated" or "separate". Consolidated
        accounting combines the financial statements of a parent company and its
        subsidiaries, while separate accounting keeps them separate. Defaults to
        'consolidated'.

        Raises:
            ValueError: If the accounting method is invalid.
        """
        return self._acc_unit

    @acc_method.setter
    def acc_method(self, value: Literal["consolidated", "separate"]):
        if value in {"consolidated", "separate"}:
            self._acc_method = value
        else:
            raise ValueError("acc_method expects 'consolidated' or 'separate'")
        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_co_df()

    @property
    def acc_unit(self) -> float:
        """Gets or sets the accounting unit for the financial statements.

        The "acc_unit" is a constant that will divide all company
        accounting values. The constant must be a number greater than
        zero or one of the following strings:
            - "thousand" to represent thousands       (1,000)
            - "million" to represent millions     (1,000,000)
            - "billion" to represent billions (1,000,000,000)

        Returns:
            The current accounting unit.

        Raises:
            ValueError: If the accounting unit is invalid.

        Examples:
            To set the accounting unit to millions:
                company.acc_unit = "million"

            To set the accounting unit to a custom factor, e.g., 10,000:
                company.acc_unit = 10_000
        """
        return self._acc_unit

    @acc_unit.setter
    def acc_unit(self, value: int | float | str) -> float | int:
        match value:
            case "thousand":
                self._acc_unit = 1_000
            case "million":
                self._acc_unit = 1_000_000
            case "billion":
                self._acc_unit = 1_000_000_000
            case str():  # Add this case to catch invalid strings
                raise ValueError("Invalid string for Accounting Unit")
            case v if v > 0:
                self._acc_unit = v
            case _:
                raise ValueError("Accounting Unit is invalid")

        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_co_df()

    @property
    def tax_rate(self) -> float:
        """Gets or sets company 'tax_rate' property.

        The "tax_rate" is used to calculate some of the company
        indicators, such as EBIT and net income. The default value
        is 0.34, which is the standard corporate tax rate in Brazil.

        Returns:
            The tax rate value.

        Raises:
            ValueError: If the tax rate is not a float between 0 and 1.

        Examples:
            To set the tax rate to 21%:
                company.tax_rate = 0.21

            To set the tax rate to 0% (tax exempt):
                company.tax_rate = 0.0
        """
        return self._tax_rate

    @tax_rate.setter
    def tax_rate(self, value: float):
        if 0 <= value <= 1:
            self._tax_rate = value
        else:
            raise ValueError("Company 'tax_rate' value is invalid")

    @property
    def language(self) -> str:
        """Gets or sets the language of the account names.

        It is the language used in the account names of the financial
        statements. This property accepts a string representing the
        desired language. Supported languages are "english" and
        "portuguese". The default is 'english'.

        Returns:
            The language used in the account names.

        Raises:
            KeyError: If the provided language is not supported.
        """

        return self._language

    @language.setter
    def language(self, language: str):
        # Supported languages
        list_languages = ["english", "portuguese"]
        if language.lower() in list_languages:
            self._language = language.capitalize()
        else:
            sup_lang = f"Supported languages: {', '.join(list_languages)}"
            raise KeyError(f"'{language}' not supported. {sup_lang}")

    def _set_co_df(self) -> pd.DataFrame:
        """Sets the company data frame.

        This method creates a data frame with the company's financial
        statements.
        """
        basic_data_types = {
            "co_name": str,
            "co_id": int,
            "co_fiscal_id": str,
            "report_type": str,
            "report_version": int,
            "period_reference": "datetime64[ns]",
            "period_begin": "datetime64[ns]",
            "period_end": "datetime64[ns]",
            "period_order": str,
            "acc_code": str,
            "acc_name": str,
            "acc_method": str,
            "acc_fixed": bool,
            "acc_value": float,
            "equity_statement_column": str,
        }
        # Create the company data frame
        co_df = (
            pd.read_pickle(cf.FINLOGIC_DF_PATH)
            .query(
                "co_id == @self._co_id and acc_method == @self._acc_method",
                engine="python",
            )
            .astype(basic_data_types)
            .sort_values(by="acc_code", ignore_index=True)
        )

        # Change acc_unit only for accounts different from 3.99
        co_df["acc_value"] = np.where(
            co_df["acc_code"].str.startswith("3.99"),
            co_df["acc_value"],
            co_df["acc_value"] / self._acc_unit,
        )

        self._name = co_df["co_name"].iloc[0]
        expr = 'report_type == "annual"'
        self._first_annual = co_df.query(expr)["period_end"].min()
        self._last_annual = co_df.query(expr)["period_end"].max()
        expr = 'report_type == "quarterly"'
        self._last_quarterly = co_df.query(expr)["period_end"].max()

        # Drop columns that are already company properties
        co_df.drop(
            columns=["co_name", "co_id", "co_fiscal_id", "acc_method"], inplace=True
        )
        # Set company data frame
        self._co_df = co_df

    def info(self) -> dict:
        """Return a dictionay with company info."""
        company_info = {
            "Name": self._name,
            "CVM ID": self._co_id,
            "Fiscal ID (CNPJ)": self._co_fiscal_id,
            "Total Accounting Rows": len(self._co_df.index),
            "Selected Tax Rate": self._tax_rate,
            "Selected Accounting Method": self._acc_method,
            "Selected Accounting Unit": self._acc_unit,
            "First Annual Report": self._first_annual.strftime("%Y-%m-%d"),
            "Last Annual Report": self._last_annual.strftime("%Y-%m-%d"),
            "Last Quarterly Report": self._last_quarterly.strftime("%Y-%m-%d"),
        }
        return company_info

    def _build_report(self, dfi: pd.DataFrame) -> pd.DataFrame:
        # keep only last quarterly fs
        if self._last_annual > self._last_quarterly:
            df = dfi.query('report_type == "annual"').copy()
            df.query(
                "period_order == 'PREVIOUS' or \
                 period_end == @self._last_annual",
                inplace=True,
            )
        else:
            df = dfi.query(
                'report_type == "annual" or \
                 period_end == @self._last_quarterly'
            ).copy()
            df.query(
                "period_order == 'PREVIOUS' or \
                 period_end == @self._last_quarterly or \
                 period_end == @self._last_annual",
                inplace=True,
            )

        # Create output dataframe with only the index
        dfo = df.sort_values(by="period_end", ascending=True)[
            ["acc_name", "acc_code", "acc_fixed"]
        ].drop_duplicates(subset="acc_code", ignore_index=True, keep="last")

        periods = list(df["period_end"].sort_values().unique())
        for period in periods:
            year_cols = ["acc_value", "acc_code"]
            df_year = df.query("period_end == @period")[year_cols].copy()
            period_str = period.strftime("%Y-%m-%d")
            if period == self._last_quarterly:
                period_str += " (ttm)"
            df_year.rename(columns={"acc_value": period_str}, inplace=True)
            dfo = pd.merge(dfo, df_year, how="left", on=["acc_code"])

        return dfo.sort_values("acc_code", ignore_index=True)

    def report(
        self,
        report_type: str,
        acc_level: int | None = None,
        num_years: int = 0,
    ) -> pd.DataFrame:
        """Generate an accounting report for the company.

        This function generates a report representing one of the financial
        statements for the company adjusted by the attributes passed.

        Args:
            report_type: Type of financial report to be generated. Options
                include: "assets", "cash", "current_assets",
                "non_current_assets", "liabilities", "debt",
                "current_liabilities", "non_current_liabilities",
                "liabilities_and_equity", "equity", "income",
                "earnings_per_share", "comprehensive_income",
                "changes_in_equity", "cash_flow" and "added_value".
            acc_level: Detail level to show for account codes. Options are 2,
                3, 4 or None. Defaults to None. How the values works:
                    2    -> X.YY       (show 2 levels)
                    3    -> X.YY.ZZ    (show 3 levels)
                    4    -> X.YY.ZZ.WW (show 4 levels)
                    None -> X...       (default: show all accounts)
            num_years: Number of years to include in the report. Defaults to 0
                (all years).

        Returns:
            pd.DataFrame: Generated financial report as a pandas DataFrame.

        Raises:
            ValueError: If some argument is invalid.
        """
        # Check input arguments.
        if acc_level not in {None, 2, 3, 4}:
            raise ValueError("acc_level expects None, 2, 3 or 4")

        df = self._co_df.copy()

        # Set language
        class MyDict(dict):
            """Custom dictionary class to return key if key is not found."""

            def __missing__(self, key):
                return "(pt) " + key

        if self._language == "English":
            _pten_dict = dict(cf.language_df.values)
            _pten_dict = MyDict(_pten_dict)
            df["acc_name"] = df["acc_name"].map(_pten_dict)

        # Filter dataframe for selected acc_level
        if acc_level:
            acc_code_limit = acc_level * 3 - 2  # noqa
            df.query("acc_code.str.len() <= @acc_code_limit", inplace=True)
        """
        Filter dataframe for selected report_type (report type)
        df['acc_code'].str[0].unique() -> [1, 2, 3, 4, 5, 6, 7]
        The first part of 'acc_code' is the report type
        Table of reports correspondence:
            1 -> Balance Sheet - Assets
            2 -> Balance Sheet - Liabilities and Shareholders’ Equity
            3 -> Income
            4 -> Comprehensive Income
            5 -> Changes in Equity
            6 -> Cash Flow (Indirect Method)
            7 -> Added Value
        """
        report_types = {
            "assets": ["1"],
            "cash": ["1.01.01", "1.01.02"],
            "current_assets": ["1.01"],
            "non_current_assets": ["1.02"],
            "liabilities": ["2.01", "2.02"],
            "debt": ["2.01.04", "2.02.01"],
            "current_liabilities": ["2.01"],
            "non_current_liabilities": ["2.02"],
            "liabilities_and_equity": ["2"],
            "equity": ["2.03"],
            "income": ["3"],
            "earnings_per_share": ["3.99"],
            "comprehensive_income": ["4"],
            "changes_in_equity": ["5"],
            "cash_flow": ["6"],
            "added_value": ["7"],
        }
        acc_codes = report_types[report_type]
        expr = ""
        for count, acc_code in enumerate(acc_codes):
            if count > 0:
                expr += " or "
            expr += f'acc_code.str.startswith("{acc_code}")'
        df.query(expr, inplace=True)

        # remove earnings per share from income statment
        if report_type == "income":
            df = df[~df["acc_code"].str.startswith("3.99")]

        if report_type in {"income", "cash_flow"}:
            df = self._calculate_ttm(df)

        df.reset_index(drop=True, inplace=True)

        report_df = self._build_report(df)
        report_df.set_index(keys="acc_code", drop=True, inplace=True)
        # Show only selected years
        if num_years > 0:
            cols = report_df.columns.to_list()
            cols = cols[0:2] + cols[-num_years:]
            report_df = report_df[cols]
        return report_df

    def _calculate_ttm(self, dfi: pd.DataFrame) -> pd.DataFrame:
        if self._last_annual > self._last_quarterly:
            return dfi.query('report_type == "annual"').copy()

        df1 = dfi.query("period_end == @self._last_quarterly").copy()
        df1.query("period_begin == period_begin.min()", inplace=True)

        df2 = dfi.query("period_reference == @self._last_quarterly").copy()
        df2.query("period_begin == period_begin.min()", inplace=True)
        df2["acc_value"] = -df2["acc_value"]

        df3 = dfi.query("period_end == @self._last_annual").copy()

        df_ttm = (
            pd.concat([df1, df2, df3], ignore_index=True)[["acc_code", "acc_value"]]
            .groupby(by="acc_code")
            .sum()
            .reset_index()
        )
        df1.drop(columns="acc_value", inplace=True)
        df_ttm = pd.merge(df1, df_ttm)
        df_ttm["report_type"] = "quarterly"
        df_ttm["period_begin"] = self._last_quarterly - pd.DateOffset(years=1)

        df_annual = dfi.query('report_type == "annual"').copy()

        return pd.concat([df_annual, df_ttm], ignore_index=True)

    def custom_report(
        self,
        acc_list: list[str],
        num_years: int = 0,
    ) -> pd.DataFrame:
        """Generate a financial report from custom list of accounting codes

        Args:
            acc_list: A list of strings containg accounting codes to be used
                in the report
            num_years: Select how many years to show in the report.
                Defaults to 0 (show all years)

        Returns:
            pd.DataFrame: Generated financial report as a pandas DataFrame.

        Raises:
            ValueError: If some argument is invalid.
        """
        df_as = self.report("assets")
        df_le = self.report("liabilities_and_equity")
        df_is = self.report("income")
        df_cf = self.report("cash_flow")
        dfo = pd.concat([df_as, df_le, df_is, df_cf]).query("acc_code == @acc_list")
        # Show only selected years
        if num_years > 0:
            cols = dfo.columns.to_list()
            cols = cols[0:2] + cols[-num_years:]
            dfo = dfo[cols]
        return dfo

    @staticmethod
    def _prior_values(s: pd.Series, is_prior: bool) -> pd.Series:
        """Shift row to the right in order to obtain series previous values"""
        if is_prior:
            arr = s.iloc[:-1].values
            return np.append(np.nan, arr)
        else:
            return s

    def indicators(self, num_years: int = 0, is_prior: bool = True) -> pd.DataFrame:
        """Calculate the company main operating indicators.

        Args:
            num_years: Number of years to consider for calculation. If 0, use
                all available years. Defaults to 0.
            is_prior: Divide return measurements by book values from the end of
                the prior year (see Damodaran reference). Defaults to True.

        Returns:
            pd.DataFrame: Dataframe containing calculated financial indicators.

        References:
            [1] Aswath Damodaran, "Return on Capital (ROC), Return on Invested
                Capital (ROIC) and Return on Equity (ROE): Measurement and
                Implications.", 2007,
                https://people.stern.nyu.edu/adamodar/pdfoles/papers/returnmeasures.pdf
                https://people.stern.nyu.edu/adamodar/New_Home_Page/datafile/variable.htm
        """
        df_as = self.report("assets")
        df_le = self.report("liabilities_and_equity")
        df_in = self.report("income")
        df_cf = self.report("cash_flow")
        df = pd.concat([df_as, df_le, df_in, df_cf]).drop(
            columns=["acc_fixed", "acc_name"]
        )
        # Calculate indicators series
        revenues = df.loc["3.01"]
        gross_profit = df.loc["3.03"]
        ebit = df.loc["3.05"]
        ebt = df.loc["3.07"]
        effective_tax = df.loc["3.08"]
        depreciation_amortization = df.loc["6.01.01.04"]
        ebitda = ebit + depreciation_amortization
        operating_cash_flow = df.loc["6.01"]
        # capex = df.loc["6.02"]
        net_income = df.loc["3.11"]
        total_assets = df.loc["1"]
        total_assets_p = self._prior_values(total_assets, is_prior)
        equity = df.loc["2.03"]
        equity_p = self._prior_values(equity, is_prior)
        total_cash = df.loc["1.01.01"] + df.loc["1.01.02"]
        current_assets = df.loc["1.01"]
        current_liabilities = df.loc["2.01"]
        working_capital = current_assets - current_liabilities
        total_debt = df.loc["2.01.04"] + df.loc["2.02.01"]
        net_debt = total_debt - total_cash
        invested_capital = total_debt + equity - total_cash
        invested_capital_p = self._prior_values(invested_capital, is_prior)

        # Output Dataframe (dfo)
        dfo = pd.DataFrame(columns=df.columns)
        dfo.loc["revenues"] = revenues
        dfo.loc["operating_cash_flow"] = operating_cash_flow
        # dfo.loc["capex"] = capex
        dfo.loc["ebitda"] = ebitda
        dfo.loc["ebit"] = ebit
        dfo.loc["ebt"] = ebt
        dfo.loc["effective_tax_rate"] = -1 * effective_tax / ebt
        dfo.loc["net_income"] = net_income
        dfo.loc["total_cash"] = total_cash
        dfo.loc["total_debt"] = total_debt
        dfo.loc["net_debt"] = net_debt
        dfo.loc["working_capital"] = working_capital
        dfo.loc["invested_capital"] = invested_capital
        dfo.loc["return_on_assets"] = ebit * (1 - self._tax_rate) / total_assets_p
        dfo.loc["return_on_capital"] = ebit * (1 - self._tax_rate) / invested_capital_p
        dfo.loc["return_on_equity"] = net_income / equity_p
        dfo.loc["gross_margin"] = gross_profit / revenues
        dfo.loc["ebitda_margin"] = ebitda / revenues
        dfo.loc["pre_tax_operating_margin"] = ebit / revenues
        dfo.loc["after_tax_operating_margin"] = ebit * (1 - self._tax_rate) / revenues
        dfo.loc["net_margin"] = net_income / revenues

        dfo.index.name = "Company Financial Indicators"
        # Show only the selected number of years
        if num_years > 0:
            dfo = dfo[dfo.columns[-num_years:]]
        return dfo



import unittest
import finlogic as fl


class TestCompany(unittest.TestCase):
    def setUp(self):
        self.petro_sep = fl.Company(9512, acc_method="separate", acc_unit="billion")
        self.petro_con = fl.Company(9512, acc_method="consolidated", acc_unit="billion")

    def test_identifier(self):
        """Test the identifier method of the Company class.""" ""
        test_cvm_id = 4170
        test_fiscal_id = "33.592.510/0001-54"
        self.petro_sep.identifier = test_cvm_id
        self.assertEqual(self.petro_sep._co_id, test_cvm_id)
        self.assertEqual(self.petro_sep._co_fiscal_id, test_fiscal_id)

        self.petro_sep.identifier = test_fiscal_id
        self.assertEqual(self.petro_sep._co_id, test_cvm_id)
        self.assertEqual(self.petro_sep._co_fiscal_id, test_fiscal_id)

        invalid_identifier = 99999999  # Use an invalid identifier here

        with self.assertRaises(
            KeyError, msg="Should raise a KeyError with an invalid identifier"
        ):
            self.petro_sep.identifier = invalid_identifier

    def test_info(self):
        """Test the info method of the Company class."""
        # Get the info
        petro_info = self.petro_sep.info()
        # Check the info
        self.assertEqual(petro_info["Name"], "PETROLEO BRASILEIRO S.A. PETROBRAS")
        self.assertEqual(petro_info["CVM ID"], 9512)
        self.assertEqual(petro_info["Fiscal ID (CNPJ)"], "33.000.167/0001-01")

    def test_report(self):
        """Test if the report method of the Company class returns the correct
        values."""
        petro_report = self.petro_con.report(report_type="assets")
        # Get Total Assets
        assets_2009 = round(petro_report.at["1", "2009-12-31"], 3)
        assets_2015 = round(petro_report.at["1", "2015-12-31"], 3)
        assets_2020 = round(petro_report.at["1", "2020-12-31"], 3)
        # Check the reports
        self.assertEqual(assets_2009, 350.419)
        self.assertEqual(assets_2015, 900.135)
        self.assertEqual(assets_2020, 987.419)

    def test_indicators(self):
        # Petro indicators (separate)
        petro_indicators_sep = self.petro_sep.indicators()
        petro_indicators_con = self.petro_con.indicators()
        # Get the indicators (rounded to 4 decimals)
        revenues_2009_sep = round(petro_indicators_sep.at["revenues", "2009-12-31"], 4)
        total_debt_2015_sep = round(
            petro_indicators_sep.at["total_debt", "2015-12-31"], 4
        )
        roic_2021_sep = round(
            petro_indicators_sep.at["return_on_capital", "2021-12-31"], 4
        )
        # Get the indicators (rounded to 4 decimals)
        revenues_2009_con = round(petro_indicators_con.at["revenues", "2009-12-31"], 4)
        total_debt_2015_con = round(
            petro_indicators_con.at["total_debt", "2015-12-31"], 4
        )
        roic_2021_con = round(
            petro_indicators_con.at["return_on_capital", "2021-12-31"], 4
        )
        # Check the indicators
        self.assertEqual(roic_2021_sep, 0.1623)
        self.assertEqual(revenues_2009_sep, 134.0339)
        self.assertEqual(total_debt_2015_sep, 305.3460)
        # Check the indicators
        self.assertEqual(roic_2021_con, 0.2176)
        self.assertEqual(revenues_2009_con, 182.8338)
        self.assertEqual(total_debt_2015_con, 493.0230)


if __name__ == "__main__":
    unittest.main()



from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import List
import zipfile as zf
from pathlib import Path
import pandas as pd
import requests
from . import config as cf

URL_DFP = "http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/DFP/DADOS/"
URL_ITR = "http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/ITR/DADOS/"

CVM_DF_PATH = cf.DATA_PATH / "cvm_df.pkl"


def get_cvm_df() -> pd.DataFrame:
    """Get CVM files metadata."""
    if CVM_DF_PATH.is_file():
        cvm_df = pd.read_pickle(CVM_DF_PATH)
    else:
        columns = ["filename", "file_size", "etag", "last_modified"]
        cvm_df = pd.DataFrame(columns=columns)
    return cvm_df


def list_urls() -> List[str]:
    """Update the CVM Portal file base.
    Urls with CVM raw files:
    http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/DFP/DADOS/
    http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/ITR/DADOS/
    Links example:
    http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/DFP/DADOS/dfp_cia_aberta_2020.zip
    http://dados.cvm.gov.br/dados/CIA_ABERTA/DOC/ITR/DADOS/itr_cia_aberta_2020.zip
    Throughout 2021, there are already DFs for the year 2022 because the
    company's social calendar may not necessarily coincide with the official
    calendar. Because of this, 2 is added to the current year (the second limit
    of the range function is exclusive)
    """
    first_year = 2010  # First year avaible at CVM Portal.
    # Next year files will appear during current year.
    last_year = pd.Timestamp.now().year + 1
    years = list(range(first_year, last_year + 1))
    first_year_itr = last_year - 3
    urls = []
    for year in years:
        filename = f"dfp_cia_aberta_{year}.zip"
        url = f"{URL_DFP}{filename}"
        urls.append(url)
        if year >= first_year_itr:
            filename = f"itr_cia_aberta_{year}.zip"
            url = f"{URL_ITR}{filename}"
            urls.append(url)
    return urls


def update_raw_file(url: str) -> Path:
    """Update raw file from CVM portal. Return a Path if file is updated."""
    cvm_filepath = Path(cf.RAW_DIR, url[-23:])  # filename = url final
    with requests.Session() as s:
        r = s.get(url, stream=True)
        if r.status_code != requests.codes.ok:
            return None

    if Path.exists(cvm_filepath):
        local_file_size = cvm_filepath.stat().st_size
    else:
        local_file_size = 0
    url_file_size = int(r.headers["Content-Length"])
    if local_file_size == url_file_size:
        # File is already updated
        return None
    with cvm_filepath.open(mode="wb") as f:
        f.write(r.content)

    # headers["Last-Modified"] -> 'Wed, 23 Jun 2021 12:19:24 GMT'
    ts_server = pd.to_datetime(
        r.headers["Last-Modified"], format="%a, %d %b %Y %H:%M:%S %Z"
    )
    # Store URL files metadata in a DataFrame
    cvm_df = get_cvm_df()
    cvm_df.loc[pd.Timestamp.now()] = [
        cvm_filepath.name,
        r.headers["Content-Length"],
        r.headers["ETag"],
        ts_server,
    ]
    cvm_df.to_pickle(CVM_DF_PATH)
    return cvm_filepath


def update_raw_files(urls: str) -> List[Path]:
    """Update local CVM raw files asynchronously."""
    with ThreadPoolExecutor() as executor:
        results = executor.map(update_raw_file, urls)
    updated_filepaths = [r for r in results if r is not None]
    return updated_filepaths


def pre_process_file(cvm_filepath: Path) -> Path:
    """Read annual file, process it, save the result and return the file path."""
    df = pd.DataFrame()
    annual_zipfile = zf.ZipFile(cvm_filepath)
    child_filenames = annual_zipfile.namelist()

    df_list = []
    for child_filename in child_filenames[1:]:
        child_file = annual_zipfile.open(child_filename)

        # Only "DT_INI_EXERC" and "COLUNA_DF" have missing values.
        child_df = pd.read_csv(
            child_file,
            sep=";",
            encoding="iso-8859-1",
            true_values=["S"],
            false_values=["N"],
        )
        # Currency column has only one value (BRL) so it is not necessary.
        child_df = child_df.drop(columns=["MOEDA"])

        # There are two types of CVM files: DFP (annual) and ITR (quarterly).
        if cvm_filepath.name.startswith("dfp"):
            child_df["report_type"] = "annual"
        else:
            child_df["report_type"] = "quarterly"

        df_list.append(child_df)

    df = pd.concat(df_list, ignore_index=True)

    # Convert string columns to categorical.
    columns = df.select_dtypes(include="object").columns
    df[columns] = df[columns].astype("category")
    return df


def format_df(df: pd.DataFrame) -> pd.DataFrame:
    """Process a raw dataframe and return a formatted dataframe."""
    columns_translation = {
        "CNPJ_CIA": "co_fiscal_id",
        "DT_REFER": "period_reference",
        "VERSAO": "report_version",
        "DENOM_CIA": "co_name",
        "CD_CVM": "co_id",
        "GRUPO_DFP": "report_group",
        "ESCALA_MOEDA": "currency_unit",
        "ORDEM_EXERC": "period_order",
        "DT_FIM_EXERC": "period_end",
        "CD_CONTA": "acc_code",
        "DS_CONTA": "acc_name",
        "VL_CONTA": "acc_value",
        "ST_CONTA_FIXA": "acc_fixed",
        "DT_INI_EXERC": "period_begin",
        "COLUNA_DF": "equity_statement_column",
    }
    df.rename(columns=columns_translation, inplace=True)

    adjust_data_types = {
        "co_id": "UInt32",  # max. value = 600_000
        "report_version": "UInt8",  # values are 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
        "period_begin": "datetime64[ns]",
        "period_end": "datetime64[ns]",
        "period_reference": "datetime64[ns]",
    }
    df = df.astype(adjust_data_types)
    # currency_unit values are ['MIL', 'UNIDADE']
    map_dict = {"UNIDADE": 1, "MIL": 1000}
    df["currency_unit"] = df["currency_unit"].map(map_dict).astype(int)

    # Do not ajust acc_value for 3.99 codes.
    df["acc_value"] = df["acc_value"].where(
        df["acc_code"].str.startswith("3.99"),
        df["acc_value"] * df["currency_unit"],
    )
    df.drop(columns=["currency_unit"], inplace=True)

    # "period_order" values are: 'ÚLTIMO', 'PENÚLTIMO'
    map_dict = {"ÚLTIMO": "LAST", "PENÚLTIMO": "PREVIOUS"}
    df["period_order"] = df["period_order"].map(map_dict)
    """
    acc_method -> Financial Statemen Type
    Consolidated and Separate Financial Statements (IAS 27/2003)
    df['GRUPO_DFP'].unique() result:
        'DF Consolidado - Balanço Patrimonial Ativo',
        'DF Consolidado - Balanço Patrimonial Passivo',
        'DF Consolidado - Demonstração das Mutações do Patrimônio Líquido',
        'DF Consolidado - Demonstração de Resultado Abrangente',
        'DF Consolidado - Demonstração de Valor Adicionado',
        'DF Consolidado - Demonstração do Fluxo de Caixa (Método Indireto)',
        'DF Consolidado - Demonstração do Resultado',
        'DF Individual - Balanço Patrimonial Ativo',
        'DF Individual - Balanço Patrimonial Passivo',
        'DF Individual - Demonstração das Mutações do Patrimônio Líquido',
        'DF Individual - Demonstração de Resultado Abrangente',
        'DF Individual - Demonstração de Valor Adicionado',
        'DF Individual - Demonstração do Fluxo de Caixa (Método Indireto)',
        'DF Individual - Demonstração do Resultado',
    Hence, with string position 3:6 we can make:
    if == 'Con' -> consolidated statement
    if == 'Ind' -> separate statement
    """
    df["acc_method"] = (
        df["report_group"].str[3:6].map({"Con": "consolidated", "Ind": "separate"})
    )
    # 'GRUPO_DFP' data can be inferred from 'acc_code'
    df.drop(columns=["report_group"], inplace=True)
    # Correct/harmonize some account texts.
    df["acc_name"].replace(to_replace=["\xa0ON\xa0", "On"], value="ON", inplace=True)
    # Remove duplicated accounts
    cols = list(df.columns)
    # cols.remove("acc_value")
    df.drop_duplicates(cols, keep="last", inplace=True)
    columns_order = [
        "co_name",
        "co_id",
        "co_fiscal_id",
        "report_type",
        "report_version",
        "period_reference",
        "period_begin",
        "period_end",
        "period_order",
        "acc_code",
        "acc_name",
        "acc_method",
        "acc_fixed",
        "acc_value",
        "equity_statement_column",
    ]
    df = df[columns_order]
    # Most values in datetime and string columns are the same.
    # So these remaining columns can be converted to category.
    columns = df.select_dtypes(include=["datetime64[ns]", "object"]).columns
    df[columns] = df[columns].astype("category")

    return df


def process_file(cvm_filepath: Path) -> Path:
    """Process the annual file and return the path to the processed file."""
    df = pre_process_file(cvm_filepath)
    df = format_df(df)
    processed_filepath = cf.PROCESSED_DIR / cvm_filepath.with_suffix(".pkl.zst").name
    df.to_pickle(processed_filepath)
    return processed_filepath


def process_files(
    workers: int, cvm_filepaths: List[Path], asynchronous: bool
) -> List[Path]:
    """
    Execute function 'pre_process_file' and return
    a list with filenames for the processed files.
    """
    if asynchronous:
        with ProcessPoolExecutor(max_workers=workers) as executor:
            results = executor.map(process_file, cvm_filepaths)
        processed_filepaths = [r for r in results]
    else:
        processed_filepaths = [
            process_file(cvm_filepath) for cvm_filepath in cvm_filepaths
        ]
    return processed_filepaths



import unittest
import finlogic as fl


class TestDatabase(unittest.TestCase):
    def test_info(self):
        """Test the info method of the Database module."""
        db_info = fl.database_info()
        self.assertEqual(db_info["First financial statement"], "2009-01-31")
        self.assertTrue(db_info["Accounting rows"] > 20_000_000)

    def test_search_company(self):
        """Test the search_company method of the Database module."""
        search_result = fl.search_company("petrobras")
        """
            co_name                            co_id  co_fiscal_id
        0   PETROBRAS DISTRIBUIDORA S/A         24295  34.274.233/0001-02
        1   PETROLEO BRASILEIRO S.A. PETROBRAS   9512  33.000.167/0001-01
        """
        # Check the results
        self.assertEqual(set(search_result["co_id"]), {9512, 24295})


if __name__ == "__main__":
    unittest.main()



"""Finlogic Database module.

This module provides functions to handle financial data from the CVM Portal. It
allows updating, processing and consolidating financial statements, as well as
searching for company names in the FinLogic Database and retrieving information
about the database itself.
"""

import os
from typing import List
from pathlib import Path
import pandas as pd
from . import config as cf
from . import cvm as cv


def get_finlogic_df() -> pd.DataFrame:
    """Returns the FinLogic Database as a Pandas DataFrame.

    Returns:
        A Pandas DataFrame containing the FinLogic Database.
    """
    # Start/load main dataframe
    if cf.FINLOGIC_DF_PATH.is_file():
        finlogic_df = pd.read_pickle(cf.FINLOGIC_DF_PATH)
    else:
        finlogic_df = pd.DataFrame()
    return finlogic_df


def build_finlogic_df():
    # Get all files in processed folder
    filepaths = list(cf.PROCESSED_DIR.glob("*.zst"))
    # Guard clause: if no raw file was update, there is nothing to consolidate
    if not filepaths:
        print("No processed files to build FinLogic Database.")
        return
    # Concatenate all processed files into a single dataframe
    finlogic_df = pd.concat(
        [pd.read_pickle(filepath) for filepath in filepaths],
        ignore_index=True,
    )
    # Most values in datetime and string columns are the same.
    # So these remaining columns can be converted to category.
    columns = finlogic_df.select_dtypes(include=["datetime64[ns]", "object"]).columns
    finlogic_df[columns] = finlogic_df[columns].astype("category")
    # Keep only the newest 'report_version' in df if values are repeated
    cols = [
        "co_id",
        "report_type",
        "period_reference",
        "report_version",
        "period_order",
        "acc_method",
        "acc_code",
    ]
    finlogic_df.sort_values(by=cols, ignore_index=True, inplace=True)
    cols = finlogic_df.columns.tolist()
    cols_remove = ["report_version", "acc_value", "acc_fixed"]
    [cols.remove(col) for col in cols_remove]
    # Ascending order --> last is the newest report_version
    finlogic_df.drop_duplicates(cols, keep="last", inplace=True)
    finlogic_df.to_pickle(cf.FINLOGIC_DF_PATH)


def determine_files_to_process(updated_raw_filepaths) -> List[Path]:
    # Get filestems from updated files
    filestems_updated = [filepath.stem for filepath in updated_raw_filepaths]
    filestems_updated.sort()
    # Get existing filestems in raw folder
    filepaths_in_raw_folder = [filepath for filepath in cf.RAW_DIR.glob("*.zip")]
    filestems_in_raw_folder = [filepath.stem for filepath in filepaths_in_raw_folder]
    filestems_in_raw_folder.sort()
    # Get existing filestems in processed folder
    #  Filename example: dfp_cia_aberta_2018.pkl.zst
    filestems_in_processed_folder = [
        filepath.stem.split(".")[0] for filepath in cf.PROCESSED_DIR.glob("*.zst")
    ]
    filestems_in_processed_folder.sort()

    filestems_not_in_processed_folder = list(
        set(filestems_in_raw_folder) - set(filestems_in_processed_folder)
    )
    filestems_to_process = list(
        set(filestems_updated) | set(filestems_not_in_processed_folder)
    )

    filepaths_to_process = [
        filepath
        for filepath in filepaths_in_raw_folder
        if filepath.stem in filestems_to_process
    ]
    filepaths_to_process.sort()
    return filepaths_to_process


def update_database(asynchronous: bool = False, cpu_usage: float = 0.75):
    """Verify changes in remote files and update them in Finlogic Database.

    Args:
        asynchronous: Generate the database by processing raw files
            asynchronously. Works only on Linux and Mac. Default is False.
        cpu_usage: A number between 0 and 1, where 1 represents 100% CPU usage.
            This argument will define the number of cpu cores used for data
            processing when function asynchronous mode is set to 'True'. Default
            is 0.75.

    Returns:
        None
    """
    # Create data folders if they do not exist.
    Path.mkdir(cf.RAW_DIR, parents=True, exist_ok=True)
    Path.mkdir(cf.PROCESSED_DIR, parents=True, exist_ok=True)

    # Define the number of cpu cores for parallel data processing.
    workers = int(os.cpu_count() * cpu_usage)
    if workers < 1:
        workers = 1
    print("Updating CVM files...")
    urls = cv.list_urls()
    # urls = urls[:1]  # Test
    updated_raw_filepaths = cv.update_raw_files(urls)
    print(f"Number of CVM files updated = {len(updated_raw_filepaths)}")
    if updated_raw_filepaths:
        print("Updated files:")
        for updated_filepath in updated_raw_filepaths:
            print(f"    {cf.CHECKMARK} {updated_filepath.stem} updated.")
    else:
        print("All files are up to date.")

    filepaths_to_process = determine_files_to_process(updated_raw_filepaths)
    print("\nProcessing raw files...")
    processed_filepaths = cv.process_files(
        workers, filepaths_to_process, asynchronous=asynchronous
    )
    for processed_filepath in processed_filepaths:
        print(f"    {cf.CHECKMARK} {processed_filepath.stem.split('.')[0]} processed.")

    print("\nBuilding FinLogic Database...")
    build_finlogic_df()
    print('Updating "language" database...')
    process_language_df()
    print(f"{cf.CHECKMARK} FinLogic database updated!")


def database_info() -> dict:
    """Returns general information about FinLogic Database.

    This function generates a dictionary containing main information about
    FinLogic Database, such as the database path, file size, last update call,
    last modified dates, size in memory, number of accounting rows, unique
    accounting codes, companies, unique financial statements, first financial
    statement date and last financial statement date.

    Returns:
        A dictionary containing the FinLogic Database information.
    """
    finlogic_df = get_finlogic_df()
    if finlogic_df.empty:
        print("Finlogic Database is empty")
        return

    cvm_df = cv.get_cvm_df()
    file_date_unix = round(cf.FINLOGIC_DF_PATH.stat().st_mtime, 0)
    memory_size = finlogic_df.memory_usage(index=True, deep=True).sum()
    statements_cols = ["co_id", "report_version", "report_type", "period_reference"]
    statements_num = len(finlogic_df.drop_duplicates(subset=statements_cols).index)
    first_statement = finlogic_df["period_end"].astype("datetime64[ns]").min()
    last_statement = finlogic_df["period_end"].astype("datetime64[ns]").max()

    info_dict = {
        "Path": cf.DATA_PATH,
        "File size (MB)": round(cf.FINLOGIC_DF_PATH.stat().st_size / 1024**2, 1),
        "Last update call": cvm_df.index.max().round("1s").isoformat(),
        "Last modified": pd.Timestamp.fromtimestamp(file_date_unix).isoformat(),
        "Last updated data": cvm_df["last_modified"].max().isoformat(),
        "Memory size (MB)": round(memory_size / 1024**2, 1),
        "Accounting rows": len(finlogic_df.index),
        "Unique accounting codes": finlogic_df["acc_code"].nunique(),
        "Number of companies": finlogic_df["co_id"].nunique(),
        "Unique financial statements": statements_num,
        "First financial statement": first_statement.strftime("%Y-%m-%d"),
        "Last financial statement": last_statement.strftime("%Y-%m-%d"),
    }

    return info_dict


def search_company(company_name: str) -> pd.DataFrame:
    """Search for a company name in FinLogic Database.

    This function searches the 'co_name' column in the FinLogic Database for
    company names that contain the provided expression. It returns a DataFrame
    containing the search results, with each row representing a unique company
    that matches the search criteria.

    Args:
        company_name (str): A string to search for in the FinLogic Database
            'co_name' column.

    Returns:
        pd.DataFrame: A DataFrame containing the search results, with columns
            'co_name', 'co_id', and 'co_fiscal_id' for each unique company that
            matches the search criteria.
    """

    company_name = company_name.upper()
    df = (
        pd.read_pickle(cf.FINLOGIC_DF_PATH)
        .query("co_name.str.contains(@company_name)", engine="python")
        .sort_values(by="co_name")
        .drop_duplicates(subset="co_id", ignore_index=True)[
            ["co_name", "co_id", "co_fiscal_id"]
        ]
    )
    return df


def process_language_df():
    """Process language dataframe."""
    language_df = pd.read_csv(cf.URL_LANGUAGE)
    Path.mkdir(cf.INTERIM_DIR, parents=True, exist_ok=True)
    language_df.to_csv(cf.LANGUAGE_DF_PATH, compression="zstd", index=False)



from pathlib import Path
import pandas as pd

base_dir = Path(__file__).parent
DATA_PATH = base_dir / "data"
FINLOGIC_DF_PATH = DATA_PATH / "finlogic_df.pkl.zst"

LANGUAGE_DF_PATH = DATA_PATH / "interim/pten_df.csv.zst"
URL_LANGUAGE = "https://raw.githubusercontent.com/fe-lipe-c/finlogic_datasets/master/data/pten_df.csv"  # noqa

RAW_DIR = DATA_PATH / "raw"
PROCESSED_DIR = DATA_PATH / "processed"
INTERIM_DIR = DATA_PATH / "interim"

CHECKMARK = "\033[32m\u2714\033[0m"

# Start/load language file data
if LANGUAGE_DF_PATH.is_file():
    language_df = pd.read_csv(LANGUAGE_DF_PATH)
else:
    language_df = pd.DataFrame()
